// ==============================================================
// Public XR-Class Vessel Control Header File.
// 
// Version: 1.0
// Release Date: 1.13.2008
//
// Copyright 2007-2008 Douglas E. Beachy
// All Rights Reserved.
//
// This software is FREEWARE and may not be sold!
//
// XRVesselCtrl.h : Main header file defining the public XR-class vessel control API.
//
// Altea Aerospace: The Future Is Now.
// http://www.dougsorbiterpage.com
// ==============================================================

#pragma once

#include "Orbitersdk.h"
#include <vector>

using namespace std;

// Use this floating point constant when implementing your ship's GetCtrlAPIVersion method; also, you should compare each vessel's API 
// version against this version when you are writing interface code.
#define THIS_XRVESSELCTRL_API_VERSION 1.0f

/*
  Here is an example of how to use the XRVesselCtrl API:

  OBJHANDLE shipHandle = oapiGetVesselByName("XR01");   // find the target vessel
  if (shipHandle == NULL)
    // handle error: target ship not found

  XRVesselCtrl *pVessel = (XRVesselCtrl *)oapiGetVesselInterface(shipHandle);  // downcast to actual structure type

  // We have access to the XR vessel now; let's check the API version first
  float apiVersion = pVessel->GetCtrlAPIVersion();
  if (apiVersion < THIS_XRVESSELCTRL_API_VERSION)     // verify that the ship's API version is >= the API version we linked with
  {
    // handle error: API too old
  }
  
  // API version is OK -- as an example, let's retrieve the ship's current status...
  XRSystemStatus status;
  pVessel->GetXRSystemStatus(status);

  // ... examine the status structure and continue...
*/

//-------------------------------------------------------------------------

// Defines XR engine IDs
enum XREngineID 
{ 
    XRE_MainLeft,  XRE_MainRight, 
    XRE_HoverFore, XRE_HoverAft,
    XRE_ScramLeft, XRE_ScramRight,
    XRE_RetroLeft, XRE_RetroRight
};

// Writable values for an engine's state.
struct XREngineStateWrite
{
    double     ThrottleLevel;   // 0 <= n <= 1.0
    double     GimbalX;         // -1.0 <= n <= 1.0 : 0 = centered horizontally (yaw)
    double     GimbalY;         // -1.0 <= n <= 1.0 : 0 = centered vertically (pitch)
    double     Balance;         // -1.0 <= n <= 1.0 : 0 = equally balanced between fore/aft or left/right engines
    bool       CenteringModeX;  // true if pitch centering mode is set
    bool       CenteringModeY;  // true if yaw centering mode is set
    bool       CenteringModeBalance; // true if balance centering mode is set
    bool       AutoMode;        // true if auto mode is set
    bool       DivergentMode;   // true if divergent mode is set
};

// Readable values for an engine's state, which include all the writable values plus the values below.  
// This allows you to do things like this:
/*
    XREngineStateRead state;
    vessel.GetEngineState(XRE_MainLeft, state);   // populate the structure
    state.ThrottleLevel = 1.0;                       // set to full throttle...
    vessel.SetEngineState(XRE_MainLeft, state);   // ...and write the new state back to the vessel
*/
struct XREngineStateRead : public XREngineStateWrite
{
    double     TSFC;            // 0 <= n
    double     FlowRate;        // kg/sec
    double     Thrust;          // kN
    double     FuelLevel;       // 0 <= n <= 1.0
    double     MaxFuelLevel;    // max fuel mass for this tank in kg

    // Some or all of these values may not be available for all engines; any unsupported value will be set to -1.
    double DiffuserTemp;        // degrees K
    double BurnerTemp;          // degrees K
    double ExhaustTemp;         // degrees K
};

//-------------------------------------------------------------------------

// Door States; a "door" has five normal states: OPEN, CLOSED, OPENING, CLOSING, and FAILED
// In addtion, a door has a 'proc' showing how far it is open or closed, from 0 (close) to 1 (open).
// Also note that XRD_AirlockChamber refers to the state of the air pressure in the airlock: 
// OPEN = pressurized, CLOSED = vacuum, OPENING = pressurizing, CLOSING = depressurizing
enum XRDoorID 
{ 
    XRD_DockingPort, XRD_ScramDoors, XRD_HoverDoors, XRD_Ladder, XRD_Gear, 
    XRD_RetroDoors, XRD_OuterAirlock, XRD_InnerAirlock, XRD_AirlockChamber, 
    XRD_CrewHatch, XRD_Radiator, XRD_Speedbrake, XRD_APU, 
    XRD_CrewElevator, XRD_PayloadBayDoors
};

enum XRDoorState { XRDS_Opening, XRDS_Open, XRDS_Closing, XRDS_Closed, XRDS_Failed, XRDS_DoorNotSupported };

//-------------------------------------------------------------------------

enum XRDamageState  { XRD_offline, XRD_online, XRD_notSupported };
enum XRWarningState { XRW_warningActive, XRW_warningInactive };

// Warning, damage, and system status: XR-class vessels support incremental damage, from 1.0 (100% working)
// to 0.0 (0% working).  For example, a wing that is 50% damaged will only provide 50% of 
// its normal lift.
//
// Some systems, such as doors, are either online or offline (working or failed), so those systems return an XRDamageState enum.
// For warning status, each system returns an XRWarningState enum.
// Damage may be manually set by editing the XR Vessel's scenario file.

struct XRSystemStatus
{
    double LeftWing;
    double RightWing;
    double LeftMainEngine;
    double RightMainEngine;
    double LeftSCRAMEngine;
    double RightSCRAMEngine;
    double ForeHoverEngine;   // these are *logical* engines
    double AftHoverEngine;
    double LeftRetroEngine;
    double RightRetroEngine;
    double ForwardLowerRCS;
    double AftUpperRCS;
    double ForwardUpperRCS;
    double AftLowerRCS;
    double ForwardStarboardRCS;
    double AftPortRCS;
    double ForwardPortRCS;
    double AftStarboardRCS;
    double OutboardUpperPortRCS;
    double OutboardLowerStarboardRCS;
    double OutboardUpperStarboardRCS;
    double OutboardLowerPortRCS;
    double AftRCS;
    double ForwardRCS;
    XRDamageState  LeftAileron;     // includes left elevator if a separate elevator surface is present
    XRDamageState  RightAileron;    // includes right elevator if a separate elevator surface is present
    XRDamageState  LandingGear;     
    XRDamageState  DockingPort;     // "nosecone" on some ships
    XRDamageState  RetroDoors;      
    XRDamageState  TopHatch;        // "crew hatch" on some ships        
    XRDamageState  Radiator;        
    XRDamageState  Speedbrake;      // "airbrake" on some ships
    XRDamageState  PayloadBayDoors;
    XRDamageState  CrewElevator;   

    // The warning states below are not persisted in the scenario file; they are constantly
    // recalculated dynamically.
    XRWarningState HullTemperatureWarning;
    XRWarningState MainFuelWarning;
    XRWarningState RCSFuelWarning;
    XRWarningState APUFuelWarning;
    XRWarningState LOXWarning;
    XRWarningState DynamicPressureWarning;
    XRWarningState CoolantWarning;
    XRWarningState MasterWarning;     // warning active if *any* other warning active
    bool           MWSLightState;     // updated as the MWS light blinks: true = lit, false = not lit
};

//-------------------------------------------------------------------------

// Define exterior lights
enum XRLight { XRL_Nav, XRL_Beacon, XRL_Strobe };

//-------------------------------------------------------------------------

// Define standard autopilot modes
enum XRStdAutopilot 
{ 
    XRSAP_KillRot, XRSAP_Prograde, XRSAP_Retrograde, 
    XRSAP_Normal, XRSAP_AntiNormal, XRSAP_LevelHorizon, XRSAP_Hover
};

// Define autopilot state return values.  Not all ships will support all autopilot modes; in
// that case, XRAPSTATE_NotSupported is returned.
enum XRAutopilotState
{
    XRAPSTATE_Engaged, XRAPSTATE_Disengaged, XRAPSTATE_NotSupported
};

// Attitude Hold modes
enum XRAttitudeHoldMode { XRAH_HoldPitch, XRAH_HoldAOA };

// Define extended autopilot status structures
struct XRAttitudeHoldState 
{
    bool                on;
    XRAttitudeHoldMode  mode;   // XRAH_HoldPitch, etc.
    double              TargetPitch;   // in degrees (this is also the target AOA if AOA mode engaged)
    double              TargetBank;    // in degrees
};

struct XRDescentHoldState
{
    bool    on;
    double  TargetDescentRate;  // in m/s; this is negative for descending and positive for ascending
    bool    AutoLandMode;       // true = ENGAGED
};

struct XRAirspeedHoldState
{
    bool    on;
    double  TargetAirspeed;   // m/s
};

//=========================================================================
// Each vessel that supports this API will extend this abstract 
// base class.  This need not be limited to only XR-class vessels; it is up
// to the vessel developer to implement each of the virtual methods below.
//=========================================================================
class XRVesselCtrl : public VESSEL2
{
public:
    XRVesselCtrl(OBJHANDLE vessel, int fmodel);

#ifdef VSEXPRESS
    // DUMMY FUNCTION TO COMPENSATE FOR VS2005 EXPRESS BUG: this must be placed here at the top of the file.
    // This will force the compiler to bump up the function pointer lookups by four bytes, generating correct code to invoke
    // the XRVesselCtrl methods.  This line *must not* be included for the full VS2005 version.
    virtual void dummy() { }
#endif

    // API version implemented by this vessel
    // Note: this must be virtual so that the correct version will be returned by each vessel instance.
    virtual float GetCtrlAPIVersion() { return THIS_XRVESSELCTRL_API_VERSION; }

    //--------------------------------------------------------------
    // These methods must be implemented by the XR vessel subclass.
    //--------------------------------------------------------------

    // Engine State: these methods return TRUE if XREngineID is valid for this ship and (for Set) the operation succeeded; otherwise, the methods do nothing.
    // Remember that not all engines support all fields in XREngineStateWrite and not all ships support all engine types in XREngineID.
    virtual bool SetEngineState(XREngineID id, const XREngineStateWrite &state) = 0;
    virtual bool GetEngineState(XREngineID id, XREngineStateRead &state) = 0;  
    
    // Door State
    virtual bool SetDoorState(XRDoorID id, XRDoorState state) = 0;            // returns TRUE if door is valid for this ship
    virtual XRDoorState GetDoorState(XRDoorID id, double *pProc = NULL) = 0;  // returns XRDS_DoorNotSupported if door does not exist for this ship; if pProc != NULL, proc is set to 0 <= n <= 1.0

    // Read the status of the XR vessel
    virtual void GetXRSystemStatus(XRSystemStatus &status) = 0;

    // Kill all autopilots
    virtual void KillAutopilots() = 0;

    // Standard Autopilot status (on/off only)
    virtual XRAutopilotState SetStandardAP(XRStdAutopilot id, bool on) = 0;  // returns the new state of the autopilot, or XRAPSTATE_NotSupported if requested autopilot not supported
    virtual XRAutopilotState GetStandardAP(XRStdAutopilot id) = 0;

    // Extended Autopilot methods
    virtual XRAutopilotState SetAttitudeHoldAP(const XRAttitudeHoldState &state) = 0;  // returns the new state of the autopilot, or XRAPSTATE_NotSupported if autopilot not supported
    virtual XRAutopilotState GetAttitudeHoldAP(XRAttitudeHoldState &state) = 0;

    virtual XRAutopilotState SetDescentHoldAP(const XRDescentHoldState &state) = 0;    // returns the new state of the autopilot, or XRAPSTATE_NotSupported if autopilot not supported
    virtual XRAutopilotState GetDescentHoldAP(XRDescentHoldState &state) = 0;

    virtual XRAutopilotState SetAirspeedHoldAP(const XRAirspeedHoldState &state) = 0;  // returns the new state of the autopilot, or XRAPSTATE_NotSupported if autopilot not supported
    virtual XRAutopilotState GetAirspeedHoldAP(XRAirspeedHoldState &state) = 0;

    // Exterior lights: true = ON, false = OFF or not supported
    virtual bool SetExteriorLight(XRLight light, bool state) = 0;  // returns true on success, false if not supported
    virtual bool GetExteriorLight(XRLight light) = 0; 
    
    // Secondary HUD mode (1-5) : 0 = OFF
    virtual bool SetSecondaryHUDMode(int modeNumber) = 0;  // Returns: true on success, false if mode is unsupported
    virtual int GetSecondaryHUDMode() = 0;
    
    // Enable/disable tertiary HUD : 0 = OFF
    virtual bool SetTertiaryHUDMode(int modeNumber) = 0;  // use modeNumber 1 if only a single mode (i.e., on/off) is supported.  Returns true on success, or false if modeNumber is invalid.
    virtual int GetTertiaryHUDMode() = 0;

    // Reset the MWS (Master Warning System) alarm; note that under certain conditions the MWS cannot be reset (e.g., after a vessel crash)
    virtual bool ResetMasterWarningAlarm() = 0;    // returns true if MWS alarm reset successfully, false if the alarm cannot be reset

    // Center-of-Gravity shift
    virtual double GetCenterOfGravity() = 0;   // 0.0 = centered; +/- max value varies by vessel
    virtual bool   ShiftCenterOfGravity(double requestedShift) = 0;  // requestedShift = requested delta in meters from the current center-of-gravity; returns true on success, false if APU offline or if shift is maxed out

    // RCS Mode
    virtual bool IsRCSDockingMode() = 0;    // returns true if RCS DOCKING mode is active, false if RCS is in NORMAL mode
    virtual bool SetRCSDockingMode(bool on) = 0;  // set or clear RCS docking mode; returns true on success, false if RCS docking mode not supported or mode switch inhibited

    // Active EVA port
    virtual bool IsElevatorEVAPortActive() = 0;  // returns true if crew elevator is the active EVA port, false if the docking port is active
    virtual bool SetElevatorEVAPortActive(bool on) = 0;  // true = crew elevator active, false = docking port active.  Returns true on success, false if crew elevator not supported.

    // Retrieve info/warning lines on status screen and tertiary HUD
    // linesOut will be populated with a list of lines on the status screen/tertiary HUD
    // Returns: # of lines in linesOut (linesOut.size())
    virtual int GetStatusScreenText(vector<string> &linesOut) = 0;

    // TODO: add payload controls

    // TODO: add resupply / refueling support later as necessary
};
